@@ -7,7 +7,4 @@ pub struct Parameters {
 
     pub pool_idle_timeout: time::Duration,
     pub request_timeout: time::Duration,
-
-    pub max_retry_count: usize,
-    pub retry_interval: time::Duration,
 }
